﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace ImageViewer
{
    public partial class MainForm : Form
    {
        /// <summary>
        /// 開いた画像ファイル名
        /// </summary>
        private string _openedFileName;
        /// <summary>
        /// Bitmapクラス（表示用）
        /// </summary>
        private Bitmap _bmp;
        /// <summary>
        /// 表示する画像の領域
        /// </summary>
        private RectangleF _srcRect;
        /// <summary>
        /// 描画元を指定する３点の座標（左上、右上、左下の順）
        /// </summary>
        private PointF[] _srcPoints = new PointF[3];
        /// <summary>
        /// 描画用Graphicsオブジェクト
        /// </summary>
        private Graphics _gPicboxGlobal = null;
        private Graphics _gPicboxLocal = null;
        /// <summary>
        /// マウスダウンフラグ
        /// </summary>
        private bool _mouseDownFlg = false;
        /// <summary>
        /// マウスをクリックした位置の保持用
        /// </summary>
        private PointF _oldPoint;
        /// <summary>
        /// アフィン変換行列
        /// </summary>
        private System.Drawing.Drawing2D.Matrix _matAffine;

        public MainForm()
        {
            InitializeComponent();
        }

        private void MainForm_Load(object sender, EventArgs e)
        {
            // ホイールイベントの追加
            this.picImageGlobal.MouseWheel
                += new System.Windows.Forms.MouseEventHandler(this.picImage_MouseWheel);

            // リサイズイベントを強制的に実行（Graphicsオブジェクトの作成のため）
            MainForm_Resize(null, null);

            // Matrixクラスの確保（単位行列が代入される）
            _matAffine = new System.Drawing.Drawing2D.Matrix();

            // コマンドラインの確認
            var cmds = System.Environment.GetCommandLineArgs();
            if (cmds.Length > 1)
            {
                OpenImageFile(cmds[1]);
            }
        }

        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            // 解放
            if (_bmp != null)
            {
                _bmp.Dispose();
            }
        }

        private void MainForm_Resize(object sender, EventArgs e)
        {
            if ((picImageGlobal.Width == 0) || (picImageGlobal.Height == 0)) return;

            // PictureBoxと同じ大きさのBitmapクラスを作成する。
            Bitmap bmpPicBoxGlobal = new Bitmap(picImageGlobal.Width, picImageGlobal.Height);
            Bitmap bmpPicBoxLocal = new Bitmap(picImageLoacl.Width, picImageLoacl.Height);
            // 空のBitmapをPictureBoxのImageに指定する。
            picImageGlobal.Image = bmpPicBoxGlobal;
            picImageLoacl.Image = bmpPicBoxLocal;
            // Graphicsオブジェクトの作成
            _gPicboxGlobal = Graphics.FromImage(picImageGlobal.Image);
            _gPicboxLocal = Graphics.FromImage(picImageLoacl.Image);

            // 補間モードの設定（このサンプルではNearestNeighborに設定）
            _gPicboxGlobal.InterpolationMode
                = System.Drawing.Drawing2D.InterpolationMode.NearestNeighbor;
            _gPicboxLocal.InterpolationMode
                = System.Drawing.Drawing2D.InterpolationMode.NearestNeighbor;
            // 画像の描画
            DrawImage();
        }

        private void mnuFileOpen_Click(object sender, EventArgs e)
        {
            // ファイルを開くダイアログの作成 
            OpenFileDialog dlg = new OpenFileDialog();
            // ファイルフィルタ 
            dlg.Filter = "画像ﾌｧｲﾙ(*.bmp,*.jpg,*.png,*.tif,*.ico)|*.bmp;*.jpg;*.png;*.tif;*.ico";
            // ダイアログの表示 （Cancelボタンがクリックされた場合は何もしない）
            if (dlg.ShowDialog() == System.Windows.Forms.DialogResult.Cancel) return;

            // 画像ファイルを開く
            OpenImageFile(dlg.FileName);
        }

        private void mnuFileExit_Click(object sender, EventArgs e)
        {
            // 終了
            this.Close();
        }

        private void picImage_MouseDown(object sender, MouseEventArgs e)
        {
            // フォーカスの設定
            //（クリックしただけではMouseWheelイベントが有効にならない）
            picImageGlobal.Focus();
            // マウスをクリックした位置の記録
            _oldPoint.X = e.X;
            _oldPoint.Y = e.Y;
            // マウスダウンフラグ
            _mouseDownFlg = true;
        }

        private void picImage_MouseMove(object sender, MouseEventArgs e)
        {
            // マウスをクリックしながら移動中のとき
            if (_mouseDownFlg == true)
            {
                // 画像の移動
                _matAffine.Translate(e.X - _oldPoint.X, e.Y - _oldPoint.Y,
                    System.Drawing.Drawing2D.MatrixOrder.Append);
                // 画像の描画
                DrawImage();

                // ポインタ位置の保持
                _oldPoint.X = e.X;
                _oldPoint.Y = e.Y;
            }

            // マウスポインタの座標表示
            DispPixelInfo(_matAffine, e.Location);
        }

        private void picImage_MouseUp(object sender, MouseEventArgs e)
        {
            // マウスダウンフラグ
            _mouseDownFlg = false;
        }

        // マウスホイールイベント
        private void picImage_MouseWheel(object sender, MouseEventArgs e)
        {
            if (e.Delta > 0)
            {
                // 拡大
                if (_matAffine.Elements[0] < 100)  // X方向の倍率を代表してチェック
                {
                    // ポインタの位置周りに拡大
                    ScaleAt(ref _matAffine, 1.5f, e.Location);
                }
            }
            else
            {
                // 縮小
                if (_matAffine.Elements[0] > 0.01)  // X方向の倍率を代表してチェック
                {
                    // ポインタの位置周りに縮小
                    ScaleAt(ref _matAffine, 1.0f / 1.5f, e.Location);
                }
            }
            // 画像の描画
            DrawImage();
        }

        private void picImage_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left)
            {
                // 画像全体を表示
                ZoomFit(ref _matAffine, _bmp, picImageGlobal);
            }
            if (e.Button == MouseButtons.Right)
            {
                // マウスポインタ周りに等倍表示
                ScaleAt(ref _matAffine, 1f / _matAffine.Elements[0], e.Location);
            }

            // 画像の描画
            DrawImage();
        }

        /// <summary>
        /// 画像ファイルを開く
        /// </summary>
        /// <param name="filename">画像ファイルのパス</param>
        private void OpenImageFile(string filename)
        {
            if (IsImageFile(filename) == false) return;

            // 画像データの確保
            if (_bmp != null)
            {
                _bmp.Dispose();
            }
            _bmp = new Bitmap(filename);

            // 画像サイズ
            lblImageInfo.Text =
                _bmp.Width.ToString() + " x " +
                _bmp.Height.ToString() + " x " +
                Bitmap.GetPixelFormatSize(_bmp.PixelFormat).ToString() + "bit";

            // 表示する画像の領域
            _srcRect = new RectangleF(-0.5f, -0.5f, _bmp.Width, _bmp.Height);
            // 描画元を指定する３点の座標（左上、右上、左下の順）
            _srcPoints[0] = new PointF(_srcRect.Left, _srcRect.Top);
            _srcPoints[1] = new PointF(_srcRect.Right, _srcRect.Top);
            _srcPoints[2] = new PointF(_srcRect.Left, _srcRect.Bottom);

            // 画像全体を表示
            ZoomFit(ref _matAffine, _bmp, picImageGlobal);
            // 画像の描画
            DrawImage();

            _openedFileName = filename;

            this.Text = System.IO.Path.GetFileName(filename) + " - ImageViewer";
        }

        /// <summary>
        /// 画像の描画
        /// </summary>
        private void DrawImage()
        {
            if (_bmp == null) return;

            _gPicboxGlobal.Clear(picImageGlobal.BackColor);
            _gPicboxLocal.Clear(picImageLoacl.BackColor);
            DrawImageGlobal(_gPicboxGlobal, _matAffine, _bmp);
            DrawImageLocal(_gPicboxLocal, _matAffine, _bmp);
            
            picImageGlobal.Refresh();
            picImageLoacl.Refresh();
        }

        private void DrawImageLocal(
            Graphics g, 
            System.Drawing.Drawing2D.Matrix mat,
            Bitmap bmp)
        {
            if (bmp == null) return;

            // 描画元の領域
            var srcRect = new RectangleF(-0.5f, -0.5f, bmp.Width, bmp.Height);

            // 描画先の座標(描画元に合わせる、左上、右上、左下の順)
            var points = new PointF[]
            {
                new PointF(srcRect.Left, srcRect.Top),
                new PointF(srcRect.Right, srcRect.Top),
                new PointF(srcRect.Left, srcRect.Bottom),
            };
            // 描画先の座標をアフィン変換で求める（変換後の座標は上書きされる）
            mat.TransformPoints(points);

            // 描画
            g.DrawImage(
                bmp,
                points,
                srcRect,
                GraphicsUnit.Pixel
                );

            // パスで線を描画
            var pathString = new System.Drawing.Drawing2D.GraphicsPath();
            // 文字の追加
            pathString.AddString(
                "ローカル変換",
                new FontFamily(System.Drawing.Text.GenericFontFamilies.Serif),
                (int)FontStyle.Regular,
                18.0f,
                points[0],
                new StringFormat()
                );
            // パスの描画
            g.FillPath(Brushes.Yellow, pathString);

            // パスで文字を描画
            var pathLine = new System.Drawing.Drawing2D.GraphicsPath();
            // 線の追加
            pathLine.AddLine(0f, 0f, bmp.Width - 1, bmp.Height - 1);
            // パスをアフィン変換
            pathLine.Transform(mat);
            // パスの描画
            g.DrawPath(Pens.Red, pathLine);
        }

        /// <summary>
        /// グローバル変換による画像の描画
        /// </summary>
        /// <param name="g"></param>
        /// <param name="mat"></param>
        /// <param name="bmp"></param>
        private void DrawImageGlobal(
            Graphics g,
            System.Drawing.Drawing2D.Matrix mat,
            Bitmap bmp)
        {
            if (bmp == null) return;

            // 描画元の領域
            var srcRect = new RectangleF(-0.5f, -0.5f, bmp.Width, bmp.Height);

            // 描画先の座標(描画元に合わせる、左上、右上、左下の順)
            var points = new PointF[]
            {
                new PointF(srcRect.Left, srcRect.Top),
                new PointF(srcRect.Right, srcRect.Top),
                new PointF(srcRect.Left, srcRect.Bottom),
            };
            // グラフィックスオブジェクトにアフィン変換行列を指定する
            g.Transform = mat;

            // 描画
            g.DrawImage(
                bmp,
                points,
                srcRect,
                GraphicsUnit.Pixel
                );

            // 画像の上に線を描画
            g.DrawLine(
                Pens.Red,
                0f, 0f, bmp.Width - 1, bmp.Height - 1);

            // 画像の上に文字を描画
            g.DrawString(
                "グローバル変換",
                new Font("Arial", 10),
                new SolidBrush(Color.Yellow),
                new Point(0, 0)
            );
        }

        /// <summary>
        /// 指定した点（point）周りの拡大縮小
        /// </summary>
        /// <param name="scale">倍率</param>
        /// <param name="point">基準点の座標</param>
        private void ScaleAt(
            ref System.Drawing.Drawing2D.Matrix mat,
            float scale,
            PointF point
            )
        {
            // 原点へ移動
            mat.Translate(-point.X, -point.Y,
                System.Drawing.Drawing2D.MatrixOrder.Append);
            // 拡大縮小
            mat.Scale(scale, scale,
                System.Drawing.Drawing2D.MatrixOrder.Append);
            // 元へ戻す
            mat.Translate(point.X, point.Y,
                System.Drawing.Drawing2D.MatrixOrder.Append);
        }

        /// <summary>
        /// 画像をピクチャボックスのサイズに合わせて全体に表示するアフィン変換行列を求める
        /// </summary>
        /// <param name="mat">アフィン変換行列</param>
        /// <param name="bmp">画像データ</param>
        /// <param name="dst">描画先のピクチャボックス</param>
        private void ZoomFit(
            ref System.Drawing.Drawing2D.Matrix mat,
            Bitmap bmp,
            PictureBox dst)
        {
            // アフィン変換行列の初期化（単位行列へ）
            mat.Reset();

            int srcWidth = bmp.Width;
            int srcHeight = bmp.Height;
            int dstWidth = dst.Width;
            int dstHeight = dst.Height;

            float scale;

            // 右下に0.5画素ずらす
            mat.Translate(0.5f, 0.5f, System.Drawing.Drawing2D.MatrixOrder.Append);

            // 縦に合わせるか？横に合わせるか？
            if (srcHeight * dstWidth > dstHeight * srcWidth)
            {
                // ピクチャボックスの縦方法に画像表示を合わせる場合
                scale = dstHeight / (float)srcHeight;
                mat.Scale(scale, scale, System.Drawing.Drawing2D.MatrixOrder.Append);
                // 中央へ平行移動
                mat.Translate((dstWidth - srcWidth * scale) / 2f, 0f, System.Drawing.Drawing2D.MatrixOrder.Append);
            }
            else
            {
                // ピクチャボックスの横方法に画像表示を合わせる場合
                scale = dstWidth / (float)srcWidth;
                mat.Scale(scale, scale, System.Drawing.Drawing2D.MatrixOrder.Append);
                // 中央へ平行移動
                mat.Translate(0f, (dstHeight - srcHeight * scale) / 2f, System.Drawing.Drawing2D.MatrixOrder.Append);
            }
        }

        /// <summary>
        /// マウスポインタの位置の画像の輝度値を表示
        /// </summary>
        /// <param name="mat">画像を表示しているアフィン変換行列</param>
        /// <param name="image">表示している画像</param>
        /// <param name="pointPictureBox">表示先のピクチャボックス</param>
        private void DispPixelInfo(
            System.Drawing.Drawing2D.Matrix mat,
            PointF pointPictureBox)
        {
            if (_bmp == null) return;

            // ピクチャボックス→画像上の座標のアフィン変換行列
            var matInvert = mat.Clone();
            matInvert.Invert();

            // 画像上の座標
            var pointImage = new PointF[] { pointPictureBox };
            matInvert.TransformPoints(pointImage);

            int picX = (int)Math.Floor(pointImage[0].X + 0.5);
            int picY = (int)Math.Floor(pointImage[0].Y + 0.5);

            // 画像上の座標表示
            lblPixelInfo.Text =
                "(" +
                picX.ToString() + ", " +
                picY.ToString() + ")";
        }

        /// <summary>
        /// 指定したファイルが画像ファイルかどうか？調べる
        /// </summary>
        /// <param name="filename">調べるファイル名</param>
        /// <returns></returns>
        private bool IsImageFile(string filename)
        {
            if (System.IO.File.Exists(filename) == false) return false;

            // ファイル形式の確認
            string ext = System.IO.Path.GetExtension(filename).ToLower();
            if (
                (ext != ".bmp") &&
                (ext != ".jpg") &&
                (ext != ".png") &&
                (ext != ".tif") &&
                (ext != ".ico")
                ) return false;

            return true;
        }
    }
}
